@@ -94,7 +94,7 @@ struct sasl_rand_s {
 static char basis_64[] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/???????????????????????????????????????????????????????????????????????????????????????????????????????????????????????????????????????????";
 
-static char index_64[128] = {
+static signed char index_64[128] = {
     -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
     -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
     -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,62, -1,-1,-1,63,
@@ -526,6 +526,7 @@ int get_fqhostname(
     struct addrinfo *result;
 
     return_value = gethostname (name, namelen);
+    name[namelen-1] = '\0'; /* insure string is always 0 terminated*/
     if (return_value != 0) {
 	return (return_value);
     }
@@ -557,7 +558,9 @@ int get_fqhostname(
 	}
     }
 
-    if (result == NULL || result->ai_canonname == NULL) {
+    if (result == NULL || result->ai_canonname == NULL
+        || strchr (result->ai_canonname, '.') == NULL
+        || strlen (result->ai_canonname) > namelen -1) {
 	freeaddrinfo (result);
         if (abort_if_no_fqdn) {
 #ifdef WIN32
@@ -573,25 +576,8 @@ int get_fqhostname(
 	}
     }
 
-    if (strchr (result->ai_canonname, '.') == NULL) {
-	freeaddrinfo (result);
-        if (abort_if_no_fqdn) {
-#ifdef WIN32
-	    WSASetLastError (WSANO_DATA);
-#elif defined(ENODATA)
-	    errno = ENODATA;
-#elif defined(EADDRNOTAVAIL)
-	    errno = EADDRNOTAVAIL;
-#endif
-	    return (-1);
-	} else {
-	    goto LOWERCASE;
-	}
-    }
-
-
-/* Do we need to check for buffer overflow and set errno? */
     strncpy (name, result->ai_canonname, namelen);
+    name[namelen-1] = '\0'; /* insure string is always 0 terminated*/
     freeaddrinfo (result);
 
 LOWERCASE:
